/*
 * Title:        bits/map_t.hpp
 * Description:  C++ header-only wrapper for C-ABI data_t type
 * $Copyright (c) 2015-2017, 2019-2020 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 * @Version:     $Id: map_t.hpp 369907 2020-03-27 12:50:29Z hdy $
 */

/* This file:
 * Contains the C++ implementation of map_t
 *
 */

#ifndef _DATAT_BITS_INCLUDE_TAG
#error Must only be included from sag_connectivity_cpp.hpp, use that instead
#endif

/**
 * A map class which implements many of the functions on std::map.
 *
 * This is an unordered map that can be named.
 *
 * We don't guarantee exact method-for-method compatibility with any of the
 * std types.
 *
 * This class gives access methods to a map from data_t to data_t.
 *
 * This is moveable, but not implicitly copyable. If you want to copy it use
 * the .copy() method. If you want to move it use std::move().
 *
 * No underlying implementation of the map is guaranteed. Modifying the
 * contents will invalidate any iterators you have and any references that you
 * have to the underlying structure.
 */
class map_t: protected sag_underlying_map_t
{
	friend class data_t;
	friend class Message;
	/** \internal. Always have at least capacity for 8 items */
	static size_t MIN_CAPACITY() { return 8; }
protected:
	typedef sag_underlying_map_table_entry_t item_t;
	typedef int64_t hash_t;
public:

	template<typename DATA>
	struct pair
	{
		hash_t hash;
		DATA first;
		DATA second;
	};

	/**
	 * Forward/reverse and const/non-const iterators are implemented using this
	 * class.
	 *
	 * It implements the std::bidirectional_iterator concept: http://en.cppreference.com/w/cpp/concept/BidirectionalIterator
	 *
	 * You must not modify the map object while iterating over it. The iterator will throw (or assert on a non-NDEBUG build)
	 * if you try deferencing it after modifying the map object.
	 */
	template<typename DATA, typename UNDERLYING, typename PAIR>
	struct _iterator: public std::iterator<std::bidirectional_iterator_tag, DATA>
	{
		friend class map_t;
		typedef _iterator<DATA, UNDERLYING, PAIR> iterator;
		typedef PAIR element_t;

		_iterator(): map(nullptr), tabledata(nullptr), capacity(0), offs(0), count(0) {}
		_iterator(const map_t *map, int64_t offs, bool stepToItem=false)
			: map(map), tabledata(map->table?map->table->table:nullptr), capacity(map->table?map->table->capacity:0), offs(offs), count(map->table?map->table->count:0)
		{
			if (!valid() && stepToItem) {
				operator++();
			}
		}

		iterator &operator--() 
		{
			if (offs > -1) {
				do {
					--offs;
				} while (!valid());
			}
			return *this; 
		}
		iterator operator--(int)
		{
			iterator tmp(*this);
			operator--();
			return tmp;
		}
		iterator &operator++()
		{
			if (offs < capacity) {
				do {
					++offs;
				} while (!valid());
			}
			return *this; 
		}
		iterator operator++(int)
		{
			iterator tmp(*this);
			operator++();
			return tmp;
		}

		DATA &key() const {
			checkmodified();
			return static_cast<DATA&>(tabledata[offs].key);
		}
		DATA &value() const {
			checkmodified();
			return static_cast<DATA&>(tabledata[offs].value);
		}

		bool operator==(const iterator &other) const 
		{
			bool b = offs == other.offs;
			return b;
		}
		bool operator!=(const iterator &other) const 
		{
			bool b = !operator==(other);
			return b;
		}
		element_t &operator*() const
		{
			checkmodified();
			return reinterpret_cast<element_t &>(tabledata[offs]);
		}
		element_t *operator->() const
		{
			checkmodified();
			return reinterpret_cast<element_t *>(tabledata + offs);
		}
	protected:
		/** \internal it's valid if it's reached the end or if it points to a real item */
		bool valid() const
		{	
			return offs >= capacity || offs < 0 || tabledata[offs].hash > 1 || SAG_DATA_EMPTY != tabledata[offs].key.tag;
		}
		void checkmodified() const {
			if (
					((map && map->table) && (int64_t(map->table->capacity) != capacity || map->table->count != count || map->table->table != tabledata))
					|| ((!map || !map->table) && (nullptr != tabledata || 0 != count))
				) { 
				if (nullptr == getenv("AP_SKIP_DATA_T_ITERATOR_CHECKS")) {
					assert(false && "Iterator invalidated by modifying map");
					throw std::runtime_error("Iterator invalidated by modifying map");
				}
			}
		}
		const map_t *map;
		UNDERLYING tabledata;
		int64_t capacity;
		int64_t offs;
		uint64_t count;
	};
	/** The type of iterators on mutable maps */
	typedef _iterator<data_t, sag_underlying_map_table_entry_t *, pair<data_t> > iterator;
	/** The type of iterators on const maps */
	typedef _iterator<const data_t, sag_underlying_map_table_entry_t const *, const pair<const data_t> > const_iterator;
	/** The type of an element in a mutable map */
	typedef pair<data_t> element;
	/** The type of an element in a const map */
	typedef pair<const data_t> const_element;
	/** The type of an element to be inserted into a map */
	typedef std::pair<data_t, data_t> insertion_t;

	/** Construct an empty map */
	map_t()
	{
		table = nullptr;
	}
	/** Construct a map with (at least) capacity for n items and optional name */
	explicit map_t(size_t n, const char* name = nullptr);
	/** Construct an empty map with a name */
	explicit map_t(const char* name) : map_t(0, name)
	{}

	/** Free the underlying map contents */
	~map_t()
	{
		clear();
	}

	/** Move assignment. Other will be left as the empty map and our previous
		contents will be freed. */
	map_t &operator=(map_t &&other)
	{
		map_t tmp;
		// put us in tmp
		swap(std::move(tmp));
		// put other in us (and by extension the null tmp in other)
		swap(std::move(other));
		return *this;
	}
	/** Move constructor. Other will be left as the empty map */
	map_t(map_t &&other)
	{
		table = nullptr;
		swap(std::move(other));
	}

	/** Initializer list construction, for example `map_t{ std::make_pair(data_t{"key1"},data_t{"val1"}), std::make_pair(data_t{"key2"},data_t{"val2"}) }` */
	map_t(std::initializer_list<std::pair<data_t,data_t>> l)
	{
		table = nullptr;
		for (auto it = l.begin(); it != l.end(); ++it) {
			insert(std::make_pair(it->first.copy(), it->second.copy()));
		}
	}

	/** Swap the contents of this map and other. */
	void swap(map_t &&other)
	{
		std::swap(table, other.table);
	}

	/** Swap the name of this map and other. */
	void swapName(map_t &&other)
	{
		if (other.table && other.table->name) {
			// Need an empty table to set the name
			if (!table) resize(0);
			std::swap(table->name, other.table->name);
		}
	}

	/** Return a deep copy of this map */
	map_t copy() const
	{
		map_t newmap(size());
		for (const_iterator it = begin(); it != end(); ++it) {
			newmap.insert(std::make_pair(it->first.copy(), it->second.copy()));
		}
		newmap.setName(getName());
		return newmap;
	}

	/** Set the optional name of this map.  nullptr to undefine */
	void setName(const char *name)
	{
		if (table && table->name) {
			sag_free_string(table->name);
			table->name = nullptr;
		}

		if (name) {
			// Need an empty table to set the name
			if (!table) resize(0);
			table->name = sag_copy_string(name);
		}
	}

	/** Get the optional name of this map.  Can return nullptr. */
	const char* getName() const
	{
		if (!table) return nullptr;
		return table->name;
	}

private:
	/** This is not implicitly copiable. Use the copy() method. */
	map_t &operator=(const map_t &other) { abort(); }
	/** This is not implicitly copiable. Use the copy() method. */
	map_t(const map_t &other) { abort(); }
public:

	/** Returns true if this map deep-equals other */
	bool operator==(const map_t &other) const
	{
		if (size() != other.size()) return false;
		if ((table && table->name) || (other.table && other.table->name)) {
			if (!(table && table->name) || !(other.table && other.table->name)) return false;
			if (0 != strcmp(table->name, other.table->name)) return false;
		}
		for (const_iterator it = begin(); it != end(); ++it) {
			const_iterator jt = other.find(it.key());
			if (jt == other.end()) return false;
			if (it.value() != jt.value()) return false;
		}
		return true;
	}

	/** Returns false if this map deep-equals other */
	bool operator!=(const map_t &other) const { return !operator==(other); }

	/* Iteration: 
	 * begin() and end() methods, both const and non-const. 
	 * Unordered, so forward iteration only
	 */
	/** Forward iterator begin */
	iterator begin() { return iterator(this, 0, true); }
	/** Forward iterator end */
	iterator end() { return iterator(this, table?table->capacity:0, false); }
	/** Forward const_iterator begin */
	const_iterator begin() const { return const_iterator(this, 0, true); }
	/** Forward const_iterator end */
	const_iterator end() const { return const_iterator(this, table?table->capacity:0, false); }
	/** Forward const_iterator begin */
	const_iterator cbegin() const { return const_iterator(this, 0, true); }
	/** Forward const_iterator end */
	const_iterator cend() const { return const_iterator(this, table?table->capacity:0, false); }

	/** Returns the number of key/value pairs in the map */
	size_t size() const { return table?table->count:0; }

	/** Returns true if the map is empty (size() == 0) */
	bool empty() const { return 0 == size(); }

	/** Return a reference to the item with the given key.
		If k is not in the map then it inserts it and returns a reference to that */
	data_t &operator[](data_t &&k)
	{
		return maybe_insert(insertion_t(std::move(k), data_t())).first->second;
	}

	/** Return a reference to the item with the given key.
		If k is not in the map then it inserts it and returns a reference to that */
	data_t &operator[](const data_t &k)
	{
		return maybe_insert(std::pair<const data_t&,const data_t&>(k, data_t())).first->second;
	}
	/** Return a reference to the item with the given key.
		If k is not in the map then throws std::runtime_error */
	const data_t &operator[](const data_t &k) const
	{
		const_iterator it = find(k);
		if (it == end()) {
			throw std::runtime_error("Element not in map");
		} else {
			return it.value();
		}
	}
	/** Searches the map for an item with the given key and returns an iterator
		to that position in the map. If it is not found will return end(). */
	iterator find(const data_t &k);

	/** Searches the map for an item with the given key and returns an iterator
		to that position in the map. If it is not found will return end(). */
	const_iterator find(const data_t &k) const;

	/**
	 * Insert a new key/value pair into the map. If the key already exists
	 * then the map is not modified and an iterator to the existing item is
	 * returned.
	 * Returns pair<iterator, bool> where the iterator points to the item with
	 * the given key and the boolean is true if the map needed to be modified
	 * with the new pair.
	 *
	 * This is a move operator from the data_ts, so they will be left in the
	 * empty state. If you do not want this, use .copy(), otherwise use std::move
	 */
	std::pair<iterator, bool> insert(insertion_t &&v);

	/**
	 * Insert a new key/value pair into the map. If the key already exists
	 * then the map is not modified and an iterator to the existing item is
	 * returned.
	 * Returns pair<iterator, bool> where the iterator points to the item with
	 * the given key and the boolean is true if the map needed to be modified
	 * with the new pair.
	 *
	 * This is a move operator from the data_ts, so they will be left in the
	 * empty state. If you do not want this, use .copy(), otherwise use std::move
	 */
	std::pair<iterator, bool> insert(data_t&& k, data_t&& v) {
		return insert(std::make_pair(std::move(k), std::move(v)));
	}

	/** \internal Do not use.
	 */
	std::pair<iterator, bool> insert(insertion_t &&v, hash_t hash);

	/**
	 * Remove the item with the specified key. Returns the number of items
	 * removed (always 1)
	 */
	size_t erase(const data_t &k);
	/**
	 * Erase the item pointed to by this iterator. If the iterator is end()
	 * this is a noop.
	 */
	void erase(iterator it);
	/**
	 * Empty the map, clean the name and free the underlying data.
	 */
	void clear();

protected:
	/** \internal
	 *
	 * Change the capacity. It will be at least MIN_CAPACITY(). If the array
	 * currently contains items they will be copied to the new underlying
	 * array.
	 */
	void resize(size_t newsize);

	/** \internal put the new data at the empty space we just found */
	void really_insert(item_t &item, insertion_t &&v, hash_t hash);
	/** \internal insert with lazy copy if we actually insert */
	std::pair<iterator, bool> maybe_insert(const std::pair<const data_t&, const data_t&> &v);
	/** \internal insert with lazy move if we actually insert */
	std::pair<iterator, bool> maybe_insert(insertion_t &&v);
	/** \internal find the item and what place it would be inserted if it is new */
	std::pair<map_t::const_iterator, map_t::const_iterator> find_for_insert(const data_t &k, hash_t hash) const;

	/** \internal Is this item empty? */
	bool empty(const item_t &item) const { return 0 == item.hash && SAG_DATA_EMPTY == item.key.tag; }
	/** \internal Is this item empty? */
	bool empty(const const_iterator::element_t &item) const { return 0 == item.hash && SAG_DATA_EMPTY == item.first.tag; }
	/** \internal Is this item a hole? */
	bool hole(const item_t &item) const { return 1 == item.hash && SAG_DATA_EMPTY == item.key.tag; }
	/** \internal Is this item a hole? */
	bool hole(const const_iterator::element_t &item) const { return 1 == item.hash && SAG_DATA_EMPTY == item.first.tag; }
	/** \internal Is this item live */
	bool live(const item_t &item) const { return SAG_DATA_EMPTY != item.key.tag || item.hash > 1; }
	/** \internal Is this item live */
	bool live(const const_iterator::element_t &item) const { return SAG_DATA_EMPTY != item.first.tag || item.hash > 1; }

	/** \internal Get the table index of this hash */
	size_t index(hash_t hash) const
	{
		if (!table) return 0;
		return table->capacity ? hash % table->capacity : 0;
	}

#ifdef _SAG_INTERNAL_UNSUPPORTED_DEBUG
public:
	SAG_CONNECTIVITY_API static size_t clashes;
	SAG_CONNECTIVITY_API static size_t inserts;
	SAG_CONNECTIVITY_API static double resizeUsage;
	SAG_CONNECTIVITY_API static size_t resizes;
	SAG_CONNECTIVITY_API static size_t max_lookup_clash;
	SAG_CONNECTIVITY_API static size_t lookups;
	SAG_CONNECTIVITY_API static size_t lookup_clashes;
	static double getInsertClashRatio() { return (double)clashes/(double)inserts; }
	static double getAverageResizeRatio() { return resizeUsage/(double)resizes; }
	static void resetCounters() { clashes = inserts = resizeUsage = resizes = 0; }
	static double getAverageLookupProbes() { return 1.0+(double)lookup_clashes/(double)lookups; }
	static size_t getMaxLookupProbes() { return 1+max_lookup_clash; }
#endif
};



