/*
 * Title:        sag_internal/data.hpp
 * Description:  Implementation of data_t, list_t, map_t et al
 * $Copyright (c) 2015-2017, 2019-2020 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 * @Version:     $Id: data.hpp 369907 2020-03-27 12:50:29Z hdy $
 */


#ifndef _SAG_INTERNAL_DATA_HPP_
#define _SAG_INTERNAL_DATA_HPP_

namespace com {
namespace softwareag {
namespace _DATAT_INTERNAL_CPP_NAMESPACE { 

template<>
struct get_underlying<map_t> : ampl::type_is<sag_underlying_map_t>{};
template<>
struct get_underlying<data_t> : ampl::type_is<sag_underlying_data_t>{};
template<>
struct get_underlying<list_t> : ampl::type_is<sag_underlying_vector_t>{};

/************************************************ data_t ************************************************/

/** Construct an empty data_t */
inline data_t::data_t()
{
	memset(reinterpret_cast<sag_underlying_data_t*>(this), 0, sizeof(sag_underlying_data_t));
	tag = SAG_DATA_EMPTY;
}

/** Construct a data_t, move the contents from another data_t.
	Other will be set to the empty data_t */
inline data_t::data_t(data_t &&other)
{
	memset(reinterpret_cast<sag_underlying_data_t*>(this), 0, sizeof(sag_underlying_data_t));
	tag = SAG_DATA_EMPTY;
	swap(std::move(other));
}

inline data_t::data_t(const char *s)
{
	if (!s) throw std::runtime_error("Cannot use a null char* as a data_t value (consider using an empty data_t() instead)");
	tag = SAG_DATA_STRING;
	string = sag_copy_string(s);
}

inline data_t::data_t(const char *s, do_not_copy_t)
{
	if (!s) throw std::runtime_error("Cannot use a null char* as a data_t value (consider using an empty data_t() instead)");
	tag = SAG_DATA_CONST_STRING;
	string = s;
}

inline data_t::data_t(const char *s, size_t n)
{
	if (!s) throw std::runtime_error("Cannot use a null char* as a data_t value (consider using an empty data_t() instead)");
	tag = SAG_DATA_STRING;
	string = sag_copy_partial_string(s, n);
}

inline data_t::data_t(const std::string &s)
{
	tag = SAG_DATA_STRING;
	string = sag_copy_string(s.c_str());
}

inline void data_t::swap(data_t &&other)
{
	std::swap(static_cast<sag_underlying_data_t&>(*this), static_cast<sag_underlying_data_t&>(other));
}

inline data_t data_t::copy() const
{
	sag_underlying_data_t u = sag_copy_data(static_cast<const sag_underlying_data_t&>(*this));
	data_t &d = static_cast<data_t&>(u);
	return std::move(d);
}

template<typename V>
inline typename V::result_type data_t::apply_visitor(const V &v) const
{
	switch (tag) {
		case SAG_DATA_EMPTY: return v();
		case SAG_DATA_BOOLEAN: return v(boolean);
		case SAG_DATA_DOUBLE: return v(fp);
		case SAG_DATA_INTEGER: return v(integer);
		case SAG_DATA_DECIMAL: return v(decimal);
		case SAG_DATA_STRING: 
		case SAG_DATA_CONST_STRING: return v(string);
		case SAG_DATA_LIST: return v(static_cast<const list_t&>(list));
		case SAG_DATA_MAP: return v(static_cast<const map_t&>(map));
		case SAG_DATA_BUFFER: return v(static_cast<const buffer_t&>(buffer));
		case SAG_DATA_CUSTOM: return v(custom);
		default:
			assert(false); // indicates a serious error that shouldn't occur
			return v();
	}
}
template<typename V>
inline typename V::result_type data_t::apply_visitor(const V &v)
{
	switch (tag) {
		case SAG_DATA_EMPTY: return v();
		case SAG_DATA_BOOLEAN: return v(boolean);
		case SAG_DATA_DOUBLE: return v(fp);
		case SAG_DATA_INTEGER: return v(integer);
		case SAG_DATA_DECIMAL: return v(decimal);
		case SAG_DATA_STRING: 
		case SAG_DATA_CONST_STRING: return v(string);
		case SAG_DATA_LIST: return v(static_cast<list_t&>(list));
		case SAG_DATA_MAP: return v(static_cast<map_t&>(map));
		case SAG_DATA_BUFFER: return v(static_cast<buffer_t&>(buffer));
		case SAG_DATA_CUSTOM: return v(custom);
		default:
			assert(false); // indicates a serious error that shouldn't occur
			return v();
	}
}
inline data_t::data_t(list_t &&d)
{
	tag = SAG_DATA_LIST;
	new(&list) list_t(std::move(d));
}
inline data_t::data_t(map_t &&d)
{
	tag = SAG_DATA_MAP;
	new(&map) map_t(std::move(d));
}
inline data_t::data_t(buffer_t &&d)
{
	tag = SAG_DATA_BUFFER;
	new(&buffer) buffer_t(std::move(d));
}

template<typename T>
inline data_t::data_t(custom_t<T> &&d)
{
	tag = SAG_DATA_CUSTOM;
	new (&custom) custom_t<int>(std::move(reinterpret_cast<custom_t<int> &>(d)));
}
inline bool data_t::operator==(const data_t &other) const
{
	if (tag != other.tag) return false;
	switch (tag) {
		case SAG_DATA_EMPTY: return true;
		case SAG_DATA_BOOLEAN: return boolean == other.boolean;
		case SAG_DATA_DOUBLE: return fp == other.fp;
		case SAG_DATA_INTEGER: return integer == other.integer;
		case SAG_DATA_DECIMAL: return decimal.d == other.decimal.d; // TODO: is this correct?
		case SAG_DATA_STRING: 
		case SAG_DATA_CONST_STRING: 
			return 0 == strcmp(string, other.string);
		case SAG_DATA_LIST: return static_cast<const list_t&>(list) == static_cast<const list_t&>(other.list);
		case SAG_DATA_MAP: return static_cast<const map_t&>(map) == static_cast<const map_t&>(other.map);
		case SAG_DATA_BUFFER: return static_cast<const buffer_t&>(buffer) == static_cast<const buffer_t&>(other.buffer);
		case SAG_DATA_CUSTOM: return false;
		default:
			assert(false); // indicates a serious error that shouldn't occur
			return false;
	}
}

inline data_t::~data_t()
{
	switch (tag) {
		// value types
		case SAG_DATA_EMPTY:
		case SAG_DATA_BOOLEAN:
		case SAG_DATA_DOUBLE:
		case SAG_DATA_INTEGER:
		case SAG_DATA_DECIMAL:
		case SAG_DATA_CONST_STRING:
			break;
		// reference types
		case SAG_DATA_STRING:
			sag_free_string(string);
			break;
		case SAG_DATA_LIST: {
				list_t tmp(std::move(static_cast<list_t&>(list)));
			} 
			break;
		case SAG_DATA_MAP: {
				map_t tmp(std::move(static_cast<map_t&>(map)));
			}
			break;
		case SAG_DATA_BUFFER: {
				buffer_t tmp(std::move(static_cast<buffer_t&>(buffer)));
			}
			break;
		case SAG_DATA_CUSTOM: {
			custom_t<int> tmp(std::move(static_cast<custom_t<int> &>(custom)));
			}
			break;
		default:
			assert(false); // indicates a serious error that shouldn't occur
	}
}



/************************************************ list_t ************************************************/

inline list_t::list_t(size_t n)
{
	if (n < MIN_CAPACITY()) n = MIN_CAPACITY();
	static_cast<sag_underlying_vector_t&>(*this) = sag_allocate_vector(n);
}

inline bool list_t::operator==(const list_t &other) const
{
	if (!other.table) return table ? 0 == table->count : true;
	if (!table) return 0 == other.table->count;
	if (table->count != other.table->count) return false;
	for (size_t i = 0; i < table->count; ++i) {
		if (static_cast<data_t&>(table->data[i]) != static_cast<data_t&>(other.table->data[i])) return false;
	}
	return true;
}

inline const data_t &list_t::operator[](size_t n) const
{ 
	if (n >= size()) throw std::runtime_error("Out of bounds");
	return static_cast<const data_t&>(table->data[n]);
}

inline data_t &list_t::operator[](size_t n)
{ 
	if (n >= size()) throw std::runtime_error("Out of bounds");
	return static_cast<data_t&>(table->data[n]);
}

inline void list_t::clear()
{
	sag_free_vector(static_cast<sag_underlying_vector_t&>(*this));
	table = nullptr;
}

inline void list_t::push_back(data_t &&d)
{
	if (!table) resize(MIN_CAPACITY());
	if (table->count == table->capacity) {
		resize(table->capacity*2);
	}
	new (&(table->data[table->count++])) data_t(std::move(d));
}

inline void list_t::pop_back()
{
	if (!table) return;
	if (!table->count) return;
	data_t tmp(std::move(static_cast<data_t&>(table->data[--table->count])));
}

inline void list_t::resize(size_t newsize)
{
	if (newsize < MIN_CAPACITY()) newsize = MIN_CAPACITY();
	if (!table) {
		static_cast<sag_underlying_vector_t&>(*this) = sag_allocate_vector(newsize);
	} else {
		static_cast<sag_underlying_vector_t&>(*this) = sag_reallocate_vector(static_cast<sag_underlying_vector_t&>(*this), newsize);
	}
}

/************************************************ map_t ************************************************/

#include <sag_internal/hashes.hpp>

inline map_t::map_t(size_t n, const char* name)
{
	if (0 == n) {
		table = nullptr;
	} else {
		if (n < MIN_CAPACITY()) n = MIN_CAPACITY();
		static_cast<sag_underlying_map_t&>(*this) = sag_allocate_map(n);
	}
	setName(name);
}

inline map_t::iterator map_t::find(const data_t &k)
{
	hashes::hash_t hash = hashes::hash(k);
	std::pair<map_t::const_iterator, map_t::const_iterator> item_hole = find_for_insert(k, hash);
	return iterator(this, item_hole.first.offs);
}

inline map_t::const_iterator map_t::find(const data_t &k) const
{
	hashes::hash_t hash = hashes::hash(k);
	std::pair<map_t::const_iterator, map_t::const_iterator> item_hole = find_for_insert(k, hash);
	return const_iterator(this, item_hole.first.offs);
}

inline size_t map_t::erase(const data_t &k)
{
	iterator it = find(k);
	if (it != end()) {
		erase(it);
	}
	return 1;
}

inline void map_t::erase(map_t::iterator it)
{
	if (it != end()) {
		assert(live(table->table[it.offs]));
		data_t ktmp(std::move(static_cast<data_t&>(table->table[it.offs].key)));
		data_t vtmp(std::move(static_cast<data_t&>(table->table[it.offs].value)));
		table->table[it.offs].hash = 1; // marks a hole
		table->count--;
		table->holes++;
	}
}

inline void map_t::clear()
{
	sag_free_map(static_cast<sag_underlying_map_t&>(*this));
	table = nullptr;
}

inline void map_t::resize(size_t newsize)
{
	if (newsize < MIN_CAPACITY()) newsize = MIN_CAPACITY();
	if (!table) {
		static_cast<sag_underlying_map_t&>(*this) = sag_allocate_map(newsize);
	} else {
		static_cast<sag_underlying_map_t&>(*this) = sag_reallocate_map(static_cast<sag_underlying_map_t&>(*this), newsize);
	}
}

#ifdef _SAG_INTERNAL_UNSUPPORTED_DEBUG
struct ToString
{
	typedef std::string result_type;
	std::string operator()() const { return "(null)"; }
	std::string operator()(int64_t i) const
	{
		return convert_to_details::integerToString(i);
	}
	std::string operator()(double i) const
	{
		return convert_to_details::doubleToString(i);
	}
	std::string operator()(const char *s) const { return s; }
	template<typename T>
	std::string operator()(const T &) const { return typeid(T).name(); }
};
#endif

inline std::pair<map_t::const_iterator, map_t::const_iterator> map_t::find_for_insert(const data_t &k, hash_t hash) const
{
	if (!table) return std::make_pair(end(), end());
	if (0 == table->capacity) return std::make_pair(end(), end());
	size_t idx = index(hash);
	const_iterator item = end();
	const_iterator firsthole = end();
#ifdef _SAG_INTERNAL_UNSUPPORTED_DEBUG
	size_t lookup_clash = 0;
#endif
	do {
		item = const_iterator(this, idx);
		if (hole(*item) && firsthole == end()) {
			firsthole = item;
		} else if (empty(*item)) {
#ifdef _SAG_INTERNAL_UNSUPPORTED_DEBUG
			if (lookup_clash) map_t::clashes ++;
#endif
			return std::make_pair(end(), firsthole == end() ? item : firsthole);
		} else if (live(*item)) {
			if (hash == item->hash) { // otherwise it's a live item that's not part of this hash chain
				if (static_cast<const data_t&>(item->first) == k) { 
#ifdef _SAG_INTERNAL_UNSUPPORTED_DEBUG
					if (lookup_clash > map_t::max_lookup_clash) map_t::max_lookup_clash = lookup_clash;
					map_t::lookup_clashes += lookup_clash;
					++map_t::lookups;
#endif
					return std::make_pair(item, end());
				}
			}
#ifdef _SAG_INTERNAL_UNSUPPORTED_DEBUG
			sag_log_at("CONNECTIVITY", SAG_LOG_TRACE, "hash for %s: %ld=%lu clashes with %s: %ld=%lu", apply_visitor(ToString(), k).c_str(), hash, index(hash), apply_visitor(ToString(), item->first).c_str(), item->hash, index(item->hash));
			++lookup_clash;
#endif
		}
		++idx;
		if (idx == table->capacity) idx = 0;
	} while (idx != index(hash));
#ifdef _SAG_INTERNAL_UNSUPPORTED_DEBUG
	if (lookup_clash) map_t::clashes ++;
#endif
	return std::make_pair(end(), firsthole);
}

inline std::pair<map_t::iterator, bool> map_t::insert(map_t::insertion_t &&v)
{
	hashes::hash_t hash = hashes::hash(v.first);
	return insert(std::forward<map_t::insertion_t>(v), hash);
}

inline std::pair<map_t::iterator, bool> map_t::insert(map_t::insertion_t &&v, hash_t hash)
{
	std::pair<map_t::const_iterator, map_t::const_iterator> item_hole = find_for_insert(v.first, hash);
	if (item_hole.first != cend()) return std::make_pair(iterator(this, item_hole.first.offs), false);
	while (item_hole.second == cend() || table->count+table->holes > 0.75*table->capacity) {
		resize(table ? (table->count+1)*2 : MIN_CAPACITY());
		item_hole = find_for_insert(v.first, hash);
	}
	really_insert(table->table[item_hole.second.offs], std::forward<map_t::insertion_t>(v), hash);
#ifdef _SAG_INTERNAL_UNSUPPORTED_DEBUG
	map_t::inserts++;
#endif
	return std::make_pair(iterator(this, item_hole.second.offs), true);
}

inline std::pair<map_t::iterator, bool> map_t::maybe_insert(map_t::insertion_t &&v)
{
#ifdef _DEBUG
	// for debugging: while it's very slow, this trips over bugs where resizing invalidates references to entries or iterators.
	resize(table ? (table->count+1)*2 : MIN_CAPACITY());
#endif
	hashes::hash_t hash = hashes::hash(v.first);
	if (v.first.empty()) throw std::runtime_error("Cannot have a null key");
	std::pair<map_t::const_iterator, map_t::const_iterator> item_hole = find_for_insert(v.first, hash);
	if (item_hole.first != cend()) return std::make_pair(iterator(this, item_hole.first.offs), false);
	while (item_hole.second == cend() || table->count+table->holes > 0.75*table->capacity) {
		resize(table ? (table->count+1)*2 : MIN_CAPACITY());
		item_hole = find_for_insert(v.first, hash);
	}
	really_insert(table->table[item_hole.second.offs], std::forward<map_t::insertion_t>(v), hash);
#ifdef _SAG_INTERNAL_UNSUPPORTED_DEBUG
	map_t::inserts++;
#endif
	return std::make_pair(iterator(this, item_hole.second.offs), true);
}


inline std::pair<map_t::iterator, bool> map_t::maybe_insert(const std::pair<const data_t&, const data_t&> &v)
{
#ifdef _DEBUG
	// for debugging: while it's very slow, this trips over bugs where resizing invalidates references to entries or iterators.
	resize(table ? (table->count+1)*2 : MIN_CAPACITY());
#endif
	hashes::hash_t hash = hashes::hash(v.first);
	std::pair<map_t::const_iterator, map_t::const_iterator> item_hole = find_for_insert(v.first, hash);
	if (item_hole.first != cend()) return std::make_pair(iterator(this, item_hole.first.offs), false);
	while (item_hole.second == cend() || table->count+table->holes > 0.75*table->capacity) {
		resize(table ? (table->count+1)*2 : MIN_CAPACITY());
		item_hole = find_for_insert(v.first, hash);
	}
	really_insert(table->table[item_hole.second.offs], std::make_pair(v.first.copy(), v.second.copy()), hash);
#ifdef _SAG_INTERNAL_UNSUPPORTED_DEBUG
	map_t::inserts++;
#endif
	return std::make_pair(iterator(this, item_hole.second.offs), true);
}

inline void map_t::really_insert(item_t &item, map_t::insertion_t &&v, map_t::hash_t hash)
{
	assert(!live(item));
	table->count++;
	item.hash = hash;
	static_cast<data_t&>(item.key) = std::move(v.first);
	static_cast<data_t&>(item.value) = std::move(v.second);
}

/************************************************ metadata_t ************************************************/

inline metadata_t metadata_t::copy() const
{
	metadata_t newmap;
	new (&newmap) map_t(map_t::copy());
	return newmap;
}

/************************************************ buffer_t ************************************************/

inline buffer_t::buffer_t(size_t n)
{
	static_cast<sag_underlying_buffer_t&>(*this) = sag_allocate_buffer(n);
}

inline buffer_t buffer_t::copy() const
{
	buffer_t newbuffer;
	static_cast<sag_underlying_buffer_t&>(newbuffer) = sag_copy_buffer(static_cast<const sag_underlying_buffer_t&>(*this));
	return newbuffer;
}

inline void buffer_t::clear()
{
	sag_free_buffer(static_cast<sag_underlying_buffer_t&>(*this));
	table = nullptr;
}

/************************************************ Message ************************************************/

inline Message::Message()
{
	new (&payload)	payload_t();
	new (&metadata) map_t();
}
inline Message::Message(payload_t &&_payload)
{
	new (&payload) payload_t(std::move(_payload));
	new (&metadata) map_t();
}
inline Message::Message(payload_t &&_payload, metadata_t &&_metadata)
{
	new (&payload) payload_t(std::move(_payload));
	new (&metadata) map_t(std::move(reinterpret_cast<map_t&>(_metadata)));
}
inline Message::Message(payload_t &&_payload, map_t &&_metadata)
{
	new (&payload) payload_t(std::move(_payload));
	new (&metadata) map_t(std::move(static_cast<map_t&>(_metadata)));
}

inline Message::~Message()
{
	payload_t ptmp(std::move(static_cast<payload_t&>(payload)));
	map_t mtmp(std::move(static_cast<map_t&>(metadata)));
}

inline Message::Message(Message &&other)
{
	new (&payload) payload_t(std::move(static_cast<payload_t&>(other.payload)));
	new (&metadata) map_t(std::move(static_cast<map_t&>(other.metadata)));
}

inline const Message::payload_t &Message::getPayload() const
{
	return static_cast<const payload_t&>(payload); 
}

inline Message &Message::putMetadataValue(data_t &&key, data_t &&value) {
	if (key.type_tag() != SAG_DATA_STRING) throw std::runtime_error("Cannot use non-string keys into metadata");
	static_cast<map_t&>(metadata)[std::move(key)] = std::move(value);
	return *this;
}

inline void Message::setPayload(Message::payload_t &&_payload)
{
	payload_t old;
	old.swap(std::move(_payload));
	static_cast<payload_t&>(payload).swap(std::move(old));
}

inline void Message::setMetadata(Message::metadata_t &&_metadata)
{
	map_t old;
	old.swap(std::move(reinterpret_cast<map_t&>(_metadata)));
	reinterpret_cast<map_t&>(metadata).swap(std::move(old));
}

inline void Message::setMetadata(Message::map_t &&_metadata)
{
	map_t old;
	old.swap(std::move(_metadata));
	static_cast<map_t&>(metadata).swap(std::move(old));
}

inline void Message::swap(Message &&other)
{
	static_cast<payload_t&>(payload).swap(std::move(static_cast<payload_t&>(other.payload)));
	static_cast<map_t&>(metadata).swap(std::move(static_cast<map_t&>(other.metadata)));
}

/************************************************ MapHelper ************************************************/

namespace maphelper_details {
	// specify a getter for each type.
	template<typename T> struct Getter;

	// these can be convertd from strings
	template<> struct Getter<int64_t> { typedef convert_to_details::_get_or_convert<int64_t> visitor; };
	template<> struct Getter<double> { typedef convert_to_details::_get_or_convert<double> visitor; };
	template<> struct Getter<bool> { typedef convert_to_details::_get_or_convert<bool> visitor; };
	// these cannot
	template<> struct Getter<const map_t&> { typedef get_details::GetVisitor<const map_t> visitor; };
	template<> struct Getter<const list_t&> { typedef get_details::GetVisitor<const list_t> visitor; };
	// this can coerce primitives to string
	template<> struct Getter<std::string> 
	{
		struct visitor: public ::com::softwareag::connectivity::const_visitor<visitor, std::string>
		{
			std::string visitString(const char *s) const { return s; }
			std::string visitInteger(int64_t i) const 
			{
				return convert_to_details::integerToString(i);
			}
			std::string visitDouble(double f) const
			{
				return convert_to_details::doubleToString(f);
			}
			std::string visitBoolean(bool b) const { return b ? "true" : "false"; }
		};
	};
} // maphelper_details

template<typename T>
inline T MapHelper::getAs(const map_t &map, const char *key)
{
	map_t::const_iterator it = map.find(key);
	// throw if not found
	if (it == map.end()) throw std::runtime_error("Could not find "+std::string(key)+" in map");
	// coerce it otherwise
	return apply_visitor(typename maphelper_details::Getter<T>::visitor(), it.value());
}

template<typename T>
inline T MapHelper::getAs(const map_t &map, const char *key, const T &def)
{
	map_t::const_iterator it = map.find(key);
	// return default if not found
	if (it == map.end()) return def;
	// coerce it otherwise
	return apply_visitor(typename maphelper_details::Getter<T>::visitor(), it.value());
}

/************************************************ MapExtractor *******************************************/

inline std::string MapExtractor::stringifyKey(const data_t& data)
{
	sag_underlying_data_t rv = sag_data_to_string(reinterpret_cast<const sag_underlying_data_t&>(data));
	data_t str(std::move(reinterpret_cast<data_t&>(rv)));
	return com::softwareag::connectivity::get<const char*>(str);
}

inline std::string MapExtractor::toString()
{
	sag_underlying_data_t d;
	d.map = reinterpret_cast<const sag_underlying_map_t&>(map);
	d.tag = SAG_DATA_MAP;
	sag_underlying_data_t rv = sag_data_to_string(d);
	data_t str(std::move(reinterpret_cast<data_t&>(rv)));
	std::string prefix(displayName + "=");
	return prefix + com::softwareag::connectivity::get<const char*>(str);
}

/************************************************ custom_t ************************************************/

template<typename T>
inline custom_t<T>::custom_t()
{
	data = nullptr;
}
template<typename T>
inline custom_t<T>::custom_t(T *payload)
{
	static_cast<sag_underlying_custom_t&>(*this) = sag_allocate_custom(payload, &this_t::deleter, &this_t::copier);
}
template<typename T>
inline custom_t<T>::~custom_t()
{
	sag_free_custom(static_cast<sag_underlying_custom_t&>(*this));
}
template<typename T>
inline custom_t<T>::custom_t(custom_t &&other)
{
	data = nullptr;
	swap(std::move(other));
}
template<typename T>
custom_t<T> &custom_t<T>::operator=(custom_t<T> &&other)
{
	this_t tmp;
	swap(std::move(tmp));
	swap(std::move(other));
	return *this;
}
template<typename T>
inline void custom_t<T>::swap(custom_t<T> &&other)
{
	std::swap(this->data, other.data);
}
template<typename T>
inline custom_t<T> custom_t<T>::copy() const
{

	custom_t<T> tmp;
	static_cast<sag_underlying_custom_t&>(tmp) = sag_copy_custom(static_cast<const sag_underlying_custom_t&>(*this));
	return tmp;
}

/************************************************ to_string ************************************************/

template<typename T>
using get_underlying_t = typename get_underlying<ampl::remove_const_t<ampl::remove_ref_t<T>>>::type;

template<typename T>
inline typename std::enable_if<get_underlying<T>::value, std::string>::type to_string(const T &t)
{
	get_underlying_t<T> underlying = reinterpret_cast<const get_underlying_t<T>&>(t);
	sag_underlying_data_t udata;
	new (&udata) data_t{ std::move(reinterpret_cast<T&>(underlying)) };
	sag_underlying_data_t rv = sag_data_to_string(udata);
	data_t str(std::move(std::move(reinterpret_cast<data_t&>(rv))));
	return com::softwareag::connectivity::get<const char*>(str);
}

}}} // com.softwareag.connectivity

#endif // _SAG_INTERNAL_DATA_HPP_
